/* simplistic Token Ring <-> Ethernet bridge in userspace based on AF_PACKET
 *
 * (C) 2022 by Harald Welte <laforge@gnumonks.org>
 * for Osmocom retronetworking project.
 *
 * SPDX-License-Identifier: GPL-2.0+
 *
 * You need to call this program with two arguments:
 *   - first: the token ring device (e.g. 'tr0')
 *   - second: the ethernet device (e.g. 'eth0')
 */

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/uio.h>
#include <netinet/in.h>

#include <linux/if.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <net/ethernet.h>



#define TR_ALEN		6
#define	AC		0x10
#define LLC_FRAME	0x40

/* Copied from libosmocore.git src/bits.c */

/* look-up table for bit-reversal within a byte. Generated using:
	int i,k;
        for (i = 0 ; i < 256 ; i++) {
                uint8_t sample = 0 ;
                for (k = 0; k<8; k++) {
                        if ( i & 1 << k ) sample |= 0x80 >>  k;
                }
                flip_table[i] = sample;
        }
 */
static const uint8_t flip_table[256] = {
	0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0, 0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
	0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8, 0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
	0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4, 0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
	0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec, 0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
	0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2, 0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
	0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea, 0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
	0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6, 0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
	0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee, 0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
	0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1, 0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
	0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9, 0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
	0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5, 0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
	0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed, 0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
	0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3, 0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
	0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb, 0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
	0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7, 0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
	0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef, 0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff,
};

struct tr_hdr {
	uint8_t ac;
	uint8_t fc;
	uint8_t daddr[TR_ALEN];
	uint8_t saddr[TR_ALEN];
} __attribute((packed));


struct bridge_state {
	struct {
		int socket;
	} tr;
	struct {
		int socket;
	} eth;
};

void osmo_revbytebits_buf(uint8_t *buf, int len)
{
	int i;

	for (i = 0; i < len; i++)
		buf[i] = flip_table[buf[i]];
}

static int devname2ifindex(const char *ifname)
{
	struct ifreq ifr;
	int sk, rc;

	sk = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
	if (sk < 0)
		return sk;

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
	ifr.ifr_name[sizeof(ifr.ifr_name)-1] = 0;

	rc = ioctl(sk, SIOCGIFINDEX, &ifr);
	close(sk);
	if (rc < 0)
		return rc;

	return ifr.ifr_ifindex;
}

static int enable_promisc(int sk, int ifindex)
{
	struct packet_mreq mreq = {0};
	int rc;

	mreq.mr_ifindex = ifindex;
	mreq.mr_type = PACKET_MR_PROMISC;

	rc = setsockopt(sk, SOL_PACKET, PACKET_ADD_MEMBERSHIP, &mreq, sizeof(mreq));
	if (rc < 0) {
		fprintf(stderr, "Unable to set promiscuous mode: %s\n", strerror(errno));
		return rc;
	}

	return 0;
}

static int open_packet_socket(int ifindex, int proto)
{
	struct sockaddr_ll addr;
	int fd, rc;

	memset(&addr, 0, sizeof(addr));
	addr.sll_family = AF_PACKET;
	addr.sll_protocol = htons(proto);
	addr.sll_ifindex = ifindex;
	/* we want only packets for _other_ hosts, not packets sent by us or received for us locally */
	addr.sll_pkttype = PACKET_OTHERHOST;

	fd = socket(AF_PACKET, SOCK_RAW, htons(proto));
	if (fd < 0) {
		fprintf(stderr, "Can not create AF_PACKET socket. Are you root or have CAP_NET_RAW?\n");
		return fd;
	}

	/* there's a race condition between the above syscall and the bind() call below,
	 * causing other packets to be received in between */

	rc = bind(fd, (struct sockaddr *)&addr, sizeof(addr));
	if (rc < 0) {
		fprintf(stderr, "Can not bind AF_PACKET socket to ifindex %d\n", ifindex);
		close(fd);
		return rc;
	}

	rc = enable_promisc(fd, ifindex);
	if (rc < 0) {
		close(fd);
		return rc;
	}

	return fd;
}

static int open_packet_socket_for_netdev(const char *ifname, int proto)
{
	int rc;

	rc = devname2ifindex(ifname);
	if (rc < 0) {
		fprintf(stderr, "Cannot find net-device '%s': %s\n", ifname, strerror(errno));
		return rc;
	}

	return open_packet_socket(rc, proto);
}



static int write_tr(int socket, const struct tr_hdr *trh, const uint8_t *payload, size_t payload_len)
{
	struct iovec iov[2] = {
		{
			.iov_base = (void *) trh,
			.iov_len = sizeof(*trh),
		}, {
			.iov_base = (void *) payload,
			.iov_len = payload_len,
		}
	};

	return writev(socket, iov, 2);
}

static int eth2tr(struct bridge_state *bst)
{
	uint8_t buf[2000];
	const struct ethhdr *ethh = (struct ethhdr *) buf;
	struct tr_hdr trh;
	int rc, ethlen;

	rc = read(bst->eth.socket, buf, sizeof(buf));
	if (rc <= 0) {
		fprintf(stderr, "Error reading from ETH: %s\n", strerror(errno));
		return rc;
	}
	ethlen = rc;

	if (ethlen < sizeof(*ethh)) {
		fprintf(stderr, "short frame from ETH\n");
		return 0;
	}

	if (ethh->h_proto != htons(ETH_P_802_2)) {
		fprintf(stderr, "TR<-ETH: Not bridging unexpected frame for proto 0x%04x\n", ntohs(ethh->h_proto));
		return 0;
	}

	trh.ac = AC;
	trh.fc = LLC_FRAME;
	memcpy(&trh.daddr, ethh->h_dest, sizeof(trh.daddr));
	osmo_revbytebits_buf(trh.daddr, TR_ALEN);
	memcpy(&trh.saddr, ethh->h_source, sizeof(trh.saddr));
	osmo_revbytebits_buf(trh.saddr, TR_ALEN);

	return write_tr(bst->tr.socket, &trh, buf + sizeof(*ethh), ethlen - sizeof(*ethh));
}


static int write_eth(int socket, const struct ethhdr *ethh, const uint8_t *payload, size_t payload_len)
{
	struct iovec iov[2] = {
		{
			.iov_base = (void *) ethh,
			.iov_len = sizeof(*ethh),
		}, {
			.iov_base = (void *) payload,
			.iov_len = payload_len,
		}
	};

	return writev(socket, iov, 2);
}

static int tr2eth(struct bridge_state *bst)
{
	uint8_t buf[2000];
	const struct tr_hdr *trh = (const struct tr_hdr *) buf;
	struct ethhdr ethh;
	int rc, trlen;

	rc = read(bst->tr.socket, buf, sizeof(buf));
	if (rc <= 0) {
		fprintf(stderr, "Error reading from TR: %s\n", strerror(errno));
		return rc;
	}
	trlen = rc;

	if (trlen < sizeof(*trh)) {
		fprintf(stderr, "short frame from TR\n");
		return 0;
	}

	if (trh->fc != LLC_FRAME) {
		fprintf(stderr, "TR->ETH: Not bridging unexpected non-LLC frame for FC 0x%02x\n", trh->fc);
		return 0;
	}

	memcpy(&ethh.h_dest, trh->daddr, sizeof(ethh.h_dest));
	osmo_revbytebits_buf(ethh.h_dest, TR_ALEN);
	memcpy(&ethh.h_source, trh->saddr, sizeof(ethh.h_source));
	osmo_revbytebits_buf(ethh.h_source, TR_ALEN);
	ethh.h_proto = htons(ETH_P_802_2);

	return write_eth(bst->eth.socket, &ethh, buf + sizeof(*trh), trlen - sizeof(*trh));
}




static int bridge_main(struct bridge_state *bst)
{
	int maxfd = 0;
	int rc;
	fd_set read_fds;

	FD_ZERO(&read_fds);

	if (bst->tr.socket > maxfd)
		maxfd = bst->tr.socket;

	if (bst->eth.socket > maxfd)
		maxfd = bst->eth.socket;

	while (1) {
		FD_SET(bst->tr.socket, &read_fds);
		FD_SET(bst->eth.socket, &read_fds);

		rc = select(maxfd+1, &read_fds, NULL, NULL, NULL);
		if (rc < 0 && errno != EAGAIN) 
			break;

		if (FD_ISSET(bst->tr.socket, &read_fds))
			tr2eth(bst);

		if (FD_ISSET(bst->eth.socket, &read_fds))
			eth2tr(bst);
	}

	return rc;
}

int main(int argc, char **argv)
{
	struct bridge_state bst;
	const char *tr_name, *eth_name;

	if (argc < 3) {
		fprintf(stderr, "You must specify both TR and ETH device\n");
		exit(2);
	}

	tr_name = argv[1];
	eth_name = argv[2];

	bst.tr.socket = open_packet_socket_for_netdev(tr_name, htons(ETH_P_ALL));
	if (bst.tr.socket < 0) {
		fprintf(stderr, "Error opening TR\n");
		exit(1);
	}

	bst.eth.socket = open_packet_socket_for_netdev(eth_name, htons(ETH_P_802_2));
	if (bst.eth.socket < 0) {
		fprintf(stderr, "Error opening ETH\n");
		exit(1);
	}

	bridge_main(&bst);

	exit(0);
}
